import REGL from "regl";
import { RenderableType } from "./common";
const commonGLSL = `
  uniform vec2 resolution;

  vec4 rangeToClip(vec2 v) {
    return vec4(2.0 * v / resolution - 1.0, 0.0, 1.0);
  }

  vec4 domainToClip(vec2 v) {
    return rangeToClip(toRange(v));
  }
`;
const DEFAULT_OPTIONS = {
    canvas: null,
    alpha: false,
};
export class CandyGraph {
    constructor(options = {}) {
        var _a;
        this.commandCache = new Map();
        this.positionBufferCache = new Map();
        this.clear = (color) => {
            this.regl.clear({ color: color });
        };
        this.render = (coords, viewport, renderable) => {
            coords.scope(coords.props(), () => {
                this.scope({
                    resolution: [viewport.width, viewport.height],
                    viewport,
                }, () => {
                    this.recursiveRender(coords, renderable);
                });
            });
        };
        this.hasPositionBuffer = (name) => {
            return this.positionBufferCache.has(name);
        };
        this.getPositionBuffer = (name) => {
            return this.positionBufferCache.get(name);
        };
        this.setPositionBuffer = (name, data) => {
            this.positionBufferCache.set(name, this.regl.buffer(data));
        };
        this.clearPositionBuffers = () => {
            this.positionBufferCache.forEach((buffer) => {
                buffer.destroy();
            });
            this.positionBufferCache.clear();
        };
        const opts = Object.assign(Object.assign({}, DEFAULT_OPTIONS), options);
        this.canvas = (_a = opts.canvas) !== null && _a !== void 0 ? _a : document.createElement("canvas");
        this.regl = REGL({
            canvas: this.canvas,
            extensions: ["angle_instanced_arrays", "oes_standard_derivatives"],
            attributes: {
                depth: false,
                alpha: opts.alpha,
                premultipliedAlpha: true,
            },
        });
        this.scope = this.regl({
            uniforms: {
                resolution: this.regl.prop("resolution"),
            },
            viewport: this.regl.prop("viewport"),
            blend: {
                enable: true,
                func: { srcRGB: "src alpha", dstRGB: "one minus src alpha", srcAlpha: "one", dstAlpha: "one minus src alpha" },
            },
            depth: {
                enable: false,
            },
        });
    }
    recursiveRender(coords, renderable) {
        if (Array.isArray(renderable)) {
            for (const element of renderable) {
                this.recursiveRender(coords, element);
            }
        }
        else {
            if (renderable.kind === RenderableType.Primitive) {
                const command = this.getCommand(coords, renderable);
                renderable.render(command);
            }
            else if (renderable.kind === RenderableType.Composite) {
                if (renderable.scope) {
                    renderable.scope(renderable.props(coords), () => {
                        this.recursiveRender(coords, renderable.children());
                    });
                }
                else {
                    this.recursiveRender(coords, renderable.children());
                }
            }
        }
    }
    destroy() {
        this.clearPositionBuffers();
        this.regl.destroy();
    }
    getCommand(coords, primitive) {
        let commands = this.commandCache.get(coords.glsl);
        if (!commands) {
            commands = new Map();
            this.commandCache.set(coords.glsl, commands);
        }
        let command = commands.get(primitive.constructor);
        if (!command) {
            command = primitive.command(coords.glsl + commonGLSL);
            commands.set(primitive.constructor, command);
        }
        return command;
    }
    copyTo(sourceViewport, destinationCanvas, destinationViewport) {
        // If we're not provided a canvas, make one.
        const dest = destinationCanvas !== null && destinationCanvas !== void 0 ? destinationCanvas : document.createElement("canvas");
        // If we're not provided a destination viewport, make one the size of the source viewport.
        destinationViewport = destinationViewport !== null && destinationViewport !== void 0 ? destinationViewport : {
            x: 0,
            y: 0,
            width: sourceViewport.width,
            height: sourceViewport.height,
        };
        // If we weren't provided a destination canvas, size it according to destinationViewport.
        if (!destinationCanvas) {
            dest.width = destinationViewport.x + destinationViewport.width;
            dest.height = destinationViewport.y + destinationViewport.height;
        }
        // Copy the canvas.
        const ctx = dest.getContext("2d");
        ctx === null || ctx === void 0 ? void 0 : ctx.drawImage(this.canvas, 0, this.canvas.height - sourceViewport.height, sourceViewport.width, sourceViewport.height, destinationViewport.x, destinationViewport.y, destinationViewport.width, destinationViewport.height);
        return dest;
    }
}
//# sourceMappingURL=candygraph.js.map