import { Primitive } from "../common";
import { createDataset } from "./dataset";
const DEFAULT_OPTIONS = {
    widths: 1.0,
    colors: [0, 0, 0, 1],
};
function getPositionBuffer(cg) {
    if (!cg.hasPositionBuffer("hLines")) {
        cg.setPositionBuffer("hLines", 
        // prettier-ignore
        [
            [0, -0.5],
            [1, -0.5],
            [1, +0.5],
            [0, -0.5],
            [1, +0.5],
            [0, +0.5],
        ]);
    }
    return cg.getPositionBuffer("hLines");
}
export function createHLines(cg, lines, options) {
    const segmentGeometry = getPositionBuffer(cg);
    return new HLines(cg.regl, segmentGeometry, lines, options);
}
export class HLines extends Primitive {
    constructor(regl, segmentGeometry, lines, options = {}) {
        super();
        this.regl = regl;
        this.segmentGeometry = segmentGeometry;
        const opts = Object.assign(Object.assign({}, DEFAULT_OPTIONS), options);
        this.lines = createDataset(regl, lines);
        this.widths = createDataset(regl, opts.widths);
        this.colors = createDataset(regl, opts.colors);
    }
    command(glsl) {
        return this.regl({
            vert: `
          precision highp float;
          attribute vec2 position;
          attribute vec3 line;
          attribute float width;
          attribute vec4 color;

          varying vec4 vColor;

          ${glsl}

          float round(float v) {
            return floor(v) + floor(2.0 * fract(v));
          }

          void main() {
            vec3 ordered = line;
            if (line.x > line.y) {
              ordered.xy = ordered.yx;
            }

            float w = max(1.0, round(width));

            vec2 p0 = toRange(ordered.xz);
            vec2 p1 = toRange(ordered.yz);

            p0.x = floor(p0.x);
            p1.x = ceil(p1.x);

            if (mod(w, 2.0) == 2.0) {
              p0.y = round(p0.y);
            } else {
              p0.y = floor(p0.y) + 0.5;
            }

            p1.y = p0.y;

            vec2 xBasis = vec2(p1.x - p0.x, 0.0);
            vec2 yBasis = vec2(0.0, 1.0);

            vec2 point = p0 + xBasis * position.x + yBasis * w * position.y;
            gl_Position = rangeToClip(point);

            vColor = color;
          }`,
            frag: `
          precision highp float;
          varying vec4 vColor;

          void main() {
            gl_FragColor = vColor;
          }`,
            attributes: {
                position: {
                    buffer: this.segmentGeometry,
                    divisor: 0,
                },
                line: {
                    buffer: this.regl.prop("line"),
                    divisor: 1,
                },
                color: {
                    buffer: this.regl.prop("color"),
                    divisor: this.regl.prop("colorDivisor"),
                },
                width: {
                    buffer: this.regl.prop("width"),
                    divisor: this.regl.prop("widthDivisor"),
                },
            },
            count: 6,
            instances: this.regl.prop("instances"),
        });
    }
    render(command) {
        const { lines, colors, widths } = this;
        const instances = lines.count(3);
        command({
            instances,
            line: lines.buffer,
            color: colors.buffer,
            colorDivisor: colors.divisor(instances, 4),
            width: widths.buffer,
            widthDivisor: widths.divisor(instances, 1),
        });
    }
    dispose() {
        this.lines.dispose();
        this.colors.dispose();
        this.widths.dispose();
    }
}
//# sourceMappingURL=hlines.js.map