import { Primitive } from "../common";
import { createDataset } from "./dataset";
const DEFAULT_OPTIONS = {
    widths: 1,
    colors: [0, 0, 0, 1],
};
function getPositionBuffer(cg) {
    if (!cg.hasPositionBuffer("lineSegments")) {
        cg.setPositionBuffer("lineSegments", [
            [0, -0.5],
            [1, -0.5],
            [1, +0.5],
            [0, -0.5],
            [1, +0.5],
            [0, +0.5],
        ]);
    }
    return cg.getPositionBuffer("lineSegments");
}
export function createLineSegments(cg, points, options) {
    const segmentGeometry = getPositionBuffer(cg);
    return new LineSegments(cg.regl, segmentGeometry, points, options);
}
export class LineSegments extends Primitive {
    constructor(regl, segmentGeometry, points, options = {}) {
        super();
        this.regl = regl;
        this.segmentGeometry = segmentGeometry;
        const opts = Object.assign(Object.assign({}, DEFAULT_OPTIONS), options);
        this.points = createDataset(regl, points);
        this.widths = createDataset(regl, opts.widths);
        this.colors = createDataset(regl, opts.colors);
    }
    command(glsl) {
        return this.regl({
            vert: `
          precision highp float;
          attribute vec2 position, pointA, pointB;
          attribute float width;
          attribute vec4 color;

          varying vec4 vColor;

          ${glsl}

          void main() {
            // Transform points A and B to screen space.
            vec2 screenA = toRange(pointA);
            vec2 screenB = toRange(pointB);

            // Calculate the basis vectors for the line in screen space.
            vec2 xBasis = screenB - screenA;
            vec2 yBasis = normalize(vec2(-xBasis.y, xBasis.x));

            // Determine the screen space point position and convert it back to clip space.
            vec2 point = screenA + xBasis * position.x + yBasis * width * position.y;
            gl_Position = rangeToClip(point);

            vColor = color;
          }`,
            frag: `
          precision highp float;

          varying vec4 vColor;

          void main() {
            gl_FragColor = vColor;
          }`,
            attributes: {
                position: {
                    buffer: this.segmentGeometry,
                    divisor: 0,
                },
                pointA: {
                    buffer: this.regl.prop("points"),
                    divisor: 1,
                    offset: Float32Array.BYTES_PER_ELEMENT * 0,
                    stride: Float32Array.BYTES_PER_ELEMENT * 4,
                },
                pointB: {
                    buffer: this.regl.prop("points"),
                    divisor: 1,
                    offset: Float32Array.BYTES_PER_ELEMENT * 2,
                    stride: Float32Array.BYTES_PER_ELEMENT * 4,
                },
                width: {
                    buffer: this.regl.prop("width"),
                    divisor: this.regl.prop("widthDivisor"),
                },
                color: {
                    buffer: this.regl.prop("color"),
                    divisor: this.regl.prop("colorDivisor"),
                },
            },
            count: 6,
            instances: this.regl.prop("instances"),
        });
    }
    render(command) {
        const { points, colors, widths } = this;
        const instances = points.count(2) / 2;
        command({
            instances,
            points: points.buffer,
            color: colors.buffer,
            width: widths.buffer,
            colorDivisor: colors.divisor(instances, 4),
            widthDivisor: widths.divisor(instances, 1),
        });
    }
    dispose() {
        this.points.dispose();
        this.colors.dispose();
        this.widths.dispose();
    }
}
//# sourceMappingURL=line-segments.js.map