import { Primitive } from "../common";
import { createDataset } from "./dataset";
const DEFAULT_OPTIONS = {
    widths: 1,
    colors: [0, 0, 0, 1],
};
function roundCapJoinGeometry(resolution) {
    const instanceRoundRound = [
        [0, -0.5, 0],
        [0, -0.5, 1],
        [0, 0.5, 1],
        [0, -0.5, 0],
        [0, 0.5, 1],
        [0, 0.5, 0],
    ];
    // Add the left cap.
    for (let step = 0; step < resolution; step++) {
        const theta0 = Math.PI / 2 + ((step + 0) * Math.PI) / resolution;
        const theta1 = Math.PI / 2 + ((step + 1) * Math.PI) / resolution;
        instanceRoundRound.push([0, 0, 0]);
        instanceRoundRound.push([0.5 * Math.cos(theta0), 0.5 * Math.sin(theta0), 0]);
        instanceRoundRound.push([0.5 * Math.cos(theta1), 0.5 * Math.sin(theta1), 0]);
    }
    // Add the right cap.
    for (let step = 0; step < resolution; step++) {
        const theta0 = (3 * Math.PI) / 2 + ((step + 0) * Math.PI) / resolution;
        const theta1 = (3 * Math.PI) / 2 + ((step + 1) * Math.PI) / resolution;
        instanceRoundRound.push([0, 0, 1]);
        instanceRoundRound.push([0.5 * Math.cos(theta0), 0.5 * Math.sin(theta0), 1]);
        instanceRoundRound.push([0.5 * Math.cos(theta1), 0.5 * Math.sin(theta1), 1]);
    }
    return instanceRoundRound;
}
// Set when the position buffer is created.
let geometryCount;
function getPositionBuffer(cg) {
    if (!cg.hasPositionBuffer("lineStrip")) {
        const geometry = roundCapJoinGeometry(16);
        geometryCount = geometry.length;
        cg.setPositionBuffer("lineStrip", geometry);
    }
    return cg.getPositionBuffer("lineStrip");
}
export function createLineStrip(cg, xs, ys, options) {
    const geometry = getPositionBuffer(cg);
    return new LineStrip(cg.regl, geometry, geometryCount, xs, ys, options);
}
export class LineStrip extends Primitive {
    constructor(regl, roundCapJoin, geometryCount, xs, ys, options = {}) {
        super();
        this.regl = regl;
        this.roundCapJoin = roundCapJoin;
        this.geometryCount = geometryCount;
        const opts = Object.assign(Object.assign({}, DEFAULT_OPTIONS), options);
        this.xs = createDataset(regl, xs);
        this.ys = createDataset(regl, ys);
        this.widths = createDataset(regl, opts.widths);
        this.colors = createDataset(regl, opts.colors);
    }
    command(glsl) {
        return this.regl({
            vert: `
      precision highp float;
      attribute vec3 position;
      attribute float ax, ay, bx, by;
      attribute float width;
      attribute vec4 color;

      varying vec4 vColor;

      ${glsl}

      void main() {
        vec2 offsetA = toRange(vec2(ax, ay));
        vec2 offsetB = toRange(vec2(bx, by));
        vec2 xBasis = normalize(offsetB - offsetA);
        vec2 yBasis = vec2(-xBasis.y, xBasis.x);
        vec2 pointA = offsetA + width * (position.x * xBasis + position.y * yBasis);
        vec2 pointB = offsetB + width * (position.x * xBasis + position.y * yBasis);
        vec2 point = mix(pointA, pointB, position.z);
        gl_Position = rangeToClip(point);
        vColor = color;
      }`,
            frag: `
      precision highp float;

      varying vec4 vColor;

      void main() {
          gl_FragColor = vColor;
      }`,
            attributes: {
                position: {
                    buffer: this.roundCapJoin,
                    divisor: 0,
                },
                ax: {
                    buffer: this.regl.prop("xs"),
                    divisor: 1,
                    offset: Float32Array.BYTES_PER_ELEMENT * 0,
                },
                ay: {
                    buffer: this.regl.prop("ys"),
                    divisor: 1,
                    offset: Float32Array.BYTES_PER_ELEMENT * 0,
                },
                bx: {
                    buffer: this.regl.prop("xs"),
                    divisor: 1,
                    offset: Float32Array.BYTES_PER_ELEMENT * 1,
                },
                by: {
                    buffer: this.regl.prop("ys"),
                    divisor: 1,
                    offset: Float32Array.BYTES_PER_ELEMENT * 1,
                },
                width: {
                    buffer: this.regl.prop("width"),
                    divisor: this.regl.prop("widthDivisor"),
                },
                color: {
                    buffer: this.regl.prop("color"),
                    divisor: this.regl.prop("colorDivisor"),
                },
            },
            count: this.geometryCount,
            instances: this.regl.prop("instances"),
        });
    }
    render(command) {
        const { xs, ys, widths, colors } = this;
        const instances = xs.count(1) - 1;
        command({
            instances,
            xs: xs.buffer,
            ys: ys.buffer,
            width: widths.buffer,
            color: colors.buffer,
            widthDivisor: widths.divisor(instances, 1),
            colorDivisor: colors.divisor(instances, 4),
        });
    }
    dispose() {
        this.xs.dispose();
        this.ys.dispose();
        this.widths.dispose();
        this.colors.dispose();
    }
}
//# sourceMappingURL=line-strip.js.map